@@ -3,7 +3,6 @@ import json
 import pytz
 import logging
 import datetime
-from uuid import uuid4
 
 from swift.common.swob import Response, wsgi_to_str
 from swift.common.utils import split_path, Timestamp
@@ -36,19 +35,15 @@ def is_container(blob):
     return len(chunks) == 2 and chunks[-1] == ''
 
 
-def is_object(level, blob):
+def is_object(blob):
     chunks = blob.name.split('/')
-    return len(chunks) >= 2 and chunks[-1] != '' and len(chunks) - 1 == level
-
-
-def is_pseudofolder(level, blob):
-    chunks = blob.name.split('/')
-    return len(chunks) > 2 and chunks[-1] == '' and len(chunks) - 2 == level
+    return len(chunks) >= 2 and chunks[-1] != ''
 
 
 def all_objects(blob):
     chunks = blob.name.split('/')
-    return (len(chunks) == 2 and chunks[-1] != '') or (len(chunks) > 2 and 'application/directory' not in blob.content_type)
+    return ((len(chunks) == 2 and chunks[-1] != '')
+        or (len(chunks) > 2 and 'application/directory' not in blob.content_type))
 
 
 def blobs_size(blob_list):
@@ -78,7 +73,6 @@ class SwiftGCPDriver(BaseDriver):
         self.headers = {
             'Content-Type': 'text/html; charset=utf-8',
             'X-Timestamp': Timestamp.now().normal,
-            'X-Trans-Id': str(uuid4()),
             'Accept-Ranges': 'bytes'
         }
 
@@ -91,7 +85,9 @@ class SwiftGCPDriver(BaseDriver):
 
         prefix = self.req.params.get('prefix')
 
-        self.prefix = prefix[:-1] if prefix else ''
+        # self.prefix = prefix[:-1] if prefix else ''
+        self.prefix = prefix if prefix else ''
+
         self.account = account.lower() if account else None
         self.container = container
         self.obj = obj
@@ -124,9 +120,14 @@ class SwiftGCPDriver(BaseDriver):
     def _json_response(self, body, status, headers={}):
         self.headers.update(headers)
         self.headers['Content-Type'] = 'application/json; charset=utf-8'
-        return Response(body=json.dumps(body), status=status,
-                        headers=HeaderKeyDict(**self.headers),
-                        request=self.req)
+        if not body:
+            return Response(status=status,
+                            headers=HeaderKeyDict(**self.headers),
+                            request=self.req)
+        else:
+            return Response(body=json.dumps(body), status=status,
+                            headers=HeaderKeyDict(**self.headers),
+                            request=self.req)
 
     def _error_response(self, error, status=500):
         self.headers['Content-Type'] = 'application/json; charset=utf-8'
@@ -140,6 +141,18 @@ class SwiftGCPDriver(BaseDriver):
             return self.req.environ['swift.authorize'](self.req)
         return self.app
 
+    def _base_options(self, allowed):
+        headers = {'Allow': ', '.join(allowed)}
+
+        # not a a CORS pre-flight request
+        if not self.req.headers.get('Origin', None):
+            return self._default_response('', 204, headers)
+
+        # CORS pre-flight headers
+        headers['Access-Control-Allow-Methods'] = ', '.join(allowed)
+
+        return self._default_response('', 204, headers)
+
     def handle_account(self):
         aresp = self._is_authorized()
         if aresp:
@@ -152,8 +165,7 @@ class SwiftGCPDriver(BaseDriver):
             return self.get_account()
 
         if self.req.method in ['POST']:
-            """POST requests for Account will be forwarded"""
-            return self.app
+            return self.post_account()
 
         if self.req.method == 'DELETE':
             return self.delete_account()
@@ -167,7 +179,7 @@ class SwiftGCPDriver(BaseDriver):
         except NotFound:
             bucket = self.client.create_bucket(
                 bucket_name, location=BUCKET_LOCATION)
-            bucket.iam_configuration.uniform_bucket_level_access_enabled = False
+            # bucket.iam_configuration.uniform_bucket_level_access_enabled = False
             bucket.patch()
             return bucket
         except Exception as err:
@@ -188,10 +200,10 @@ class SwiftGCPDriver(BaseDriver):
             'X-Account-Bytes-Used': labels.get('bytes-used', 0)
         }
 
-        account_meta = self.account_info.get('meta')
-        for key in account_meta:
-            new_key = 'X-Account-Meta-{}'.format(key)
-            headers[new_key] = account_meta[key]
+        for key in labels.keys():
+            if key.find('account-meta-') >= 0:
+                new_key = 'X-{}'.format(key)
+                headers[new_key] = labels[key]
 
         return self._default_response('', 204, headers)
 
@@ -201,11 +213,28 @@ class SwiftGCPDriver(BaseDriver):
         if not account_bucket:
             return self._error_response('Get Account Error.')
 
-        account_blobs, containers = [], []
+        marker = self.req.params.get('marker')
+        end_marker = self.req.params.get('end_marker')
+        limit = self.req.params.get('limit')
+        params = {}
+
+        if marker:
+            params['start_offset'] = marker
+
+        if end_marker:
+            params['end_offset'] = end_marker
+
+        containers = []
 
         try:
-            account_blobs = list(account_bucket.list_blobs())
-            containers = filter(is_container, account_blobs)
+            account_blobs = account_bucket.list_blobs(**params)
+            for index, blob in enumerate(account_blobs):
+                if marker and index == 0:  # start_offset is inclusive
+                    continue
+                if is_container(blob):
+                    containers.append(blob)
+                if limit and (len(containers) >= int(limit)):
+                    break
         except Exception as err:
             log.error(err)
             return self._error_response(err)
@@ -221,29 +250,68 @@ class SwiftGCPDriver(BaseDriver):
             })
 
         labels = account_bucket.labels
+        container_count = labels.get('container-count', 0)
 
         headers = {
-            'X-Account-Container-Count': labels.get('container-count', 0),
+            'X-Account-Container-Count': container_count,
             'X-Account-Object-Count': labels.get('object-count', 0),
             'X-Account-Bytes-Used': labels.get('bytes-used', 0)
         }
 
-        account_meta = self.account_info.get('meta')
-        for key in account_meta:
-            new_key = 'X-Account-Meta-{}'.format(key)
-            headers[new_key] = account_meta[key]
+        for key in labels.keys():
+            if key.find('account-meta-') >= 0:
+                new_key = 'X-{}'.format(key)
+                headers[new_key] = labels[key]
 
         status = 200
-        if self.req.params.get('marker'):  # TODO: pagination
-            container_list = []
+        if len(container_list) == 0:
+            headers['Content-Length'] = 0
+            container_list = None
             status = 204
 
         return self._json_response(container_list, status, headers)
 
+    def post_account(self):
+        account_bucket = self._get_or_create_bucket(self.account)
+
+        if not account_bucket:
+            return self._error_response('Get Account Error.')
+
+        labels = account_bucket.labels
+
+        for item in self.req.headers.iteritems():
+            key, value = item
+            key = key.lower()
+            prefix = key.split('x-account-meta-')
+
+            if len(prefix) > 1:
+                meta = 'account-meta-{}'.format(prefix[1].lower())
+
+                if len(value.strip()) == 0:
+                    if meta in list(labels.keys()):
+                        del labels[meta]
+                else:
+                    labels[meta] = value.encode('utf-8')
+                continue
+
+            prefix = key.split('x-remove-account-meta-')
+
+            if len(prefix) > 1:
+                meta = 'account-meta-{}'.format(prefix[1].lower())
+
+                if meta in list(labels.keys()):
+                    del labels[meta]
+                continue
+
+        account_bucket.labels = labels
+        account_bucket.patch()
+
+        return self._default_response('', 204)
+
     def delete_account(self):
         try:
             account_bucket = self.client.get_bucket(
-                bucket_name,
+                self.account,
                 timeout=30
             )
             account_bucket.delete()
@@ -252,9 +320,12 @@ class SwiftGCPDriver(BaseDriver):
         except Conflict:
             return self._error_response('Account must be empty.')
 
-        return self._default_response('', 204, headers)
+        return self._default_response('', 204)
 
     def handle_container(self):
+        if self.req.method == 'OPTIONS':
+            return self.options_container(self.req)
+
         aresp = self._is_authorized()
         if aresp:
             return aresp
@@ -274,6 +345,11 @@ class SwiftGCPDriver(BaseDriver):
         if self.req.method == 'DELETE':
             return self.delete_container(self.req)
 
+    @cors_validation
+    def options_container(self, req, bucket=None, obj=None):
+        allowed = ['OPTIONS', 'HEAD', 'GET', 'PUT', 'POST', 'DELETE']
+        return self._base_options(allowed)
+
     @cors_validation
     def head_container(self, req, bucket=None, obj=None):
         try:
@@ -301,33 +377,76 @@ class SwiftGCPDriver(BaseDriver):
 
     @cors_validation
     def get_container(self, req, bucket=None, obj=None):
+        marker = req.params.get('marker')
+        end_marker = req.params.get('end_marker')
+        limit = req.params.get('limit')
+        delimiter = req.params.get('delimiter')
+        prefix = self.container + '/'
+        if self.prefix:
+            prefix = '/'.join([self.container, self.prefix + ('' if self.prefix[-1] == '/' else '/')])
+        params = {'prefix': prefix}
+        level = 0
+
+        if delimiter:
+            params['delimiter'] = delimiter
+            params['include_trailing_delimiter'] = True
+
+        if marker:
+            params['start_offset'] = marker
+
+        if end_marker:
+            params['end_offset'] = end_marker
+
+        if limit:
+            params['max_results'] = int(limit) + 1  # includes the folder as "container"
+
         try:
             if not bucket:
-                bucket = self.client.get_bucket(
-                    self.account,
-                    timeout=30
-                )
-            prefix = '/'.join([self.container, self.prefix])
+                bucket = self.client.get_bucket(self.account, timeout=30)
+
+            blobs = list(bucket.list_blobs(**params))
+
+            if marker:
+                blobs = blobs[1:]  # start_offset is inclusive
+
             blob = bucket.get_blob(prefix)
-            blobs = list(bucket.list_blobs(prefix=prefix))
-            level = len(prefix[:-1].split('/'))
-            pseudofolders = filter(lambda x: is_pseudofolder(level, x), blobs)
-            objects = filter(lambda x: is_object(level, x), blobs)
+
+            if blob:
+                level = len(blob.name.split('/')) - 1
+
+            objects = filter(lambda x: is_object(x), blobs)
         except Exception as err:
             log.error(err)
             return self._error_response(err)
 
         object_list = []
-        for item in (objects + pseudofolders):
-            object_list.append({
-                'name': item.name.replace(self.container + '/', ''),
-                'bytes': item.size,
-                'hash': item.md5_hash,
-                'content_type': item.content_type,
-                'last_modified': item.updated.isoformat()
-            })
+
+        for item in blobs:
+            if 'application/directory' in item.content_type:
+                has_object = len(list(filter(lambda x: item.name in x.name, objects)))
+                if item.name != prefix and has_object == 0:
+                    object_list.append({
+                        'subdir': '/'.join(item.name.split('/')[1:])
+                    })
+                elif level > 1:
+                    object_list.append({
+                        'name': item.name.replace(self.container + '/', ''),
+                        'bytes': item.size,
+                        'hash': item.md5_hash,
+                        'content_type': item.content_type,
+                        'last_modified': item.updated.isoformat()
+                    })
+            else:
+                object_list.append({
+                    'name': item.name.replace(self.container + '/', ''),
+                    'bytes': item.size,
+                    'hash': item.md5_hash,
+                    'content_type': item.content_type,
+                    'last_modified': item.updated.isoformat()
+                })
 
         headers = {}
+
         if blob and blob.metadata:
             for key, value in blob.metadata.items():
                 if key.lower() not in RESERVED_META:
@@ -336,8 +455,9 @@ class SwiftGCPDriver(BaseDriver):
                     headers[key] = value
 
         status = 200
-        if self.req.params.get('marker'):  # TODO: pagination
-            container_list = []
+        if len(object_list) == 0:
+            headers['Content-Length'] = 0
+            object_list = None
             status = 204
 
         return self._json_response(object_list, status, headers)
@@ -409,7 +529,7 @@ class SwiftGCPDriver(BaseDriver):
         except NotFound:
             bucket = self.client.create_bucket(
                 self.account, location=BUCKET_LOCATION)
-            bucket.iam_configuration.uniform_bucket_level_access_enabled = False
+            # bucket.iam_configuration.uniform_bucket_level_access_enabled = False
             bucket.patch()
         except Exception as err:
             log.error(err)
@@ -493,6 +613,9 @@ class SwiftGCPDriver(BaseDriver):
         return self._default_response('', 204)
 
     def handle_object(self):
+        if self.req.method == 'OPTIONS':
+            return self.options_object(self.req)
+
         if self.req.method == 'HEAD':
             return self.head_object(self.req)
 
@@ -586,6 +709,11 @@ class SwiftGCPDriver(BaseDriver):
 
         return updated, blob
 
+    @cors_validation
+    def options_object(self, req, bucket=None, obj=None):
+        allowed = ['OPTIONS', 'HEAD', 'GET', 'PUT', 'POST', 'DELETE']
+        return self._base_options(allowed)
+
     @cors_validation
     def head_object(self, req, bucket=None, obj=None):
         if not bucket:
@@ -608,6 +736,7 @@ class SwiftGCPDriver(BaseDriver):
                 return self._default_response('', 401)
 
         headers = self.get_object_headers(blob)
+        headers['Content-Length'] = 0
 
         return self._default_response('', 204, headers)
 
@@ -701,6 +830,8 @@ class SwiftGCPDriver(BaseDriver):
                          account_bucket,
                          container_blob,
                          bytes_used,
+                         has_obj,
+                         obj_size,
                          remove=False):
         labels = account_bucket.labels or {}
         metadata = container_blob.metadata or {}
@@ -711,15 +842,24 @@ class SwiftGCPDriver(BaseDriver):
         container_bytes_used = int(metadata.get('bytes-used', 0))
 
         if remove:
-            labels['object-count'] = max(0, account_obj_count - 1)
-            labels['bytes-used'] = max(0, account_bytes_used - bytes_used)
-            metadata['object-count'] = max(0, container_obj_count - 1)
-            metadata['bytes-used'] = max(0, container_bytes_used - bytes_used)
+            count = 1 if has_obj else 0
+            used = bytes_used if has_obj else 0
+
+            labels['object-count'] = max(0, account_obj_count - count)
+            labels['bytes-used'] = max(0, account_bytes_used - used)
+            metadata['object-count'] = max(0, container_obj_count - count)
+            metadata['bytes-used'] = max(0, container_bytes_used - used)
         else:
-            labels['object-count'] = account_obj_count + 1
-            labels['bytes-used'] = account_bytes_used + bytes_used
-            metadata['object-count'] = container_obj_count + 1
-            metadata['bytes-used'] = container_bytes_used + bytes_used
+            count = 1 if not has_obj else 0
+            used = bytes_used
+
+            if has_obj:
+                used = bytes_used - obj_size
+
+            labels['object-count'] = account_obj_count + count
+            labels['bytes-used'] = account_bytes_used + used
+            metadata['object-count'] = container_obj_count + count
+            metadata['bytes-used'] = container_bytes_used + used
 
         account_bucket.labels = labels
         container_blob.metadata = metadata
@@ -746,6 +886,14 @@ class SwiftGCPDriver(BaseDriver):
         delete_at = self.req.headers.get('x-delete-at')
         delete_after = self.req.headers.get('x-delete-after')
 
+        has_obj = False
+        obj_size = 0
+        has_blob = bucket.get_blob(obj_path)
+
+        if has_blob and has_blob.exists():
+            has_obj = True
+            obj_size = has_blob.size
+
         _, blob = self.update_object_headers(blob)
 
         if delete_at or delete_after:
@@ -769,12 +917,38 @@ class SwiftGCPDriver(BaseDriver):
                 break
             obj_data += chunk
 
+        objs =  self.obj.split('/')
+        path = self.container
+
+        is_folder = len(obj_data) == 0
+
+        if not is_folder:
+            objs = objs[:-1]
+
+        for obj in objs:
+            if not obj:
+                continue
+            path += '/' + obj
+            folder = bucket.blob(path + '/')
+            folder.upload_from_string('',
+                content_type='application/directory',
+                num_retries=3,
+                timeout=30
+            )
+
+        if is_folder:
+            headers = {
+                'Content-Type': 'application/directory',
+                'Content-Length': 0
+            }
+            return self._default_response('', 201, headers)
+
         blob.upload_from_string(obj_data, content_type=content_type)
 
         headers = self.get_object_headers(blob)
         headers['Content-Length'] = 0
 
-        self._update_counters(bucket, container_blob, len(obj_data))
+        self._update_counters(bucket, container_blob, len(obj_data), has_obj, obj_size)
 
         return self._default_response('', 201, headers)
 
@@ -818,6 +992,9 @@ class SwiftGCPDriver(BaseDriver):
         if not blob or not blob.exists():
             return self._default_response('', 404)
 
+        has_obj = True
+        obj_size = blob.size
+
         headers = self.get_object_headers(blob)
         delete_at = headers.get('x-delete-at')
 
@@ -830,6 +1007,6 @@ class SwiftGCPDriver(BaseDriver):
 
         blob.delete()
 
-        self._update_counters(bucket, container_blob, blob.size, remove=True)
+        self._update_counters(bucket, container_blob, blob.size, has_obj, obj_size, remove=True)
 
         return self._default_response('', 204)
